/*
  written by izumi@ieee.org, Oct 2016.
*/

/*
  gcc -O -Wall -g winclient.c -mwindows -lws2_32 -lmswsock -ladvapi32 -lwinmm
*/


#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdlib.h>
#include <stdio.h>
#include <w32api/mmsystem.h>

#define HOST_PORT "8765"
#define HOST_ADDR "192.168.1.11"

#define SERVO_DEFAULT   150000
#define SERVO_MIN       120000
#define SERVO_MAX       180000

#define MOTOR_DEFAULT   0
#define MOTOR_LOW       20000
#define MOTOR_MAX       65535

#define MOTOR_FWDL      0x0001
#define MOTOR_BCKL      0x0003
#define MOTOR_FWDR      0x0300
#define MOTOR_BCKR      0x0100
#define MOTOR_SYNC      0x0080 //???

#define JoypadMargin 4000

struct data_packet {
  UINT32 ctrl;
  UINT32 pwmA;
  UINT32 pwmB;
  UINT32 syncA;
  UINT32 syncB;
  UINT32 servoA;
  UINT32 servoB;
  UINT32 led;
};

void print_help(void){
  printf("\n");
  printf("Control for Keyboard\n");
  printf("q,Q,x,X ... quit\n");
  printf("f,F     ... run forward\n");
  printf("b,B     ... run backward\n");
  printf("l,L     ... turn left\n");
  printf("r,R     ... turn right\n");
  printf("s,S     ... down/up camera\n");
  printf("t,T     ... left/right camera\n");
  printf("j,J     ... use Joypad\n");
  printf("\n");
  printf("Control for Jeypad\n");
  printf("sticks  ... run/turn\n");
  printf("1,2,3,4 ... move camera\n");
  printf("5       ... use Keyboard\n");
  printf("6       ... quit\n");
  printf("\n");
}

void print_data_packet(struct data_packet *data){
  printf("ctrl %04x ",    data->ctrl);
  printf("pwm %04x %04x ",data->pwmA, data->pwmB);
  printf("syn %1x %1x ",  data->syncA, data->syncB);
  printf("srv %06x %06x ",data->servoA, data->servoB);
  printf("led %x\n",    data->led);
}

void reset_data_packet(struct data_packet *data){
  data->pwmA = 0;
  data->pwmB = 0;
  data->ctrl = 0;
  data->syncA = 0;
  data->syncB = 0;
  data->servoA = SERVO_DEFAULT;
  data->servoB = SERVO_DEFAULT;
  data->led = 0;
}

int main(int argc, char **argv) {
  WSADATA wsadata;
  SOCKET ConnectSocket=INVALID_SOCKET;
  struct addrinfo *result=NULL,*ptr=NULL,hints;
  char *hostaddr;
  struct data_packet data;
  int CtrlIF=0; // 0 ... keyboard, 1 ... JoyPad
  int count=0,sleeptime=0;
  int done=0,r=0;
  char input;

  JOYINFOEX joypad;
  int joypad_X0=0x8000,joypad_Y0=0x8000,joypad_Z0=0x8000,joypad_R0=0x8000;
  int joypad_Y=0,joypad_R=0;
  float ratio;

  printf("usage: %s [server addr]\n", argv[0]);
  print_help();
  if (argc<2) {
    hostaddr = HOST_ADDR;
  } else if (argc==2) {
    hostaddr = argv[1];
  } else {
    printf("usage: %s [server addr]\n", argv[0]);
    print_help();
    return 1;
  }

  // Initialize JoyPad
  joypad.dwSize = sizeof(JOYINFOEX);
  joypad.dwFlags = JOY_RETURNALL;
  if(JOYERR_NOERROR == joyGetPosEx(0, &joypad)) {
    printf("Controlled by JoyPad0\n");
    CtrlIF=1;
    joypad_X0=joypad.dwXpos;
    joypad_Y0=joypad.dwYpos;
    joypad_Z0=joypad.dwZpos;
    joypad_R0=joypad.dwRpos;
    printf("Center Values of (X,Y,Z,R)=(%08x,%08x,%08x,%08x)\n",
	   joypad_X0,joypad_Y0,joypad_Z0,joypad_R0);
  } else {
    print_help();
    printf("Controlled by Keyboard\n");
    CtrlIF=0;
  }
  
  // Initialize Winsock
  r=WSAStartup(MAKEWORD(2,2),&wsadata);
  if (r!=0) {
    printf("WSAStartup failed with error: %d\n",r);
    return 1;
  }
  
  ZeroMemory(&hints,sizeof(hints));
  hints.ai_family  =AF_UNSPEC;
  hints.ai_socktype=SOCK_STREAM;
  hints.ai_protocol=IPPROTO_TCP;

  // Resolve the server address and port
  r=getaddrinfo(hostaddr,HOST_PORT,&hints,&result);
  if (r!=0) {
    printf("getaddrinfo failed with error: %d\n",r);
    WSACleanup();
    return 1;
  }

  // Attempt to connect
  for(ptr=result; ptr != NULL ;ptr=ptr->ai_next) {
    fprintf(stderr,".");
    // Create a SOCKET for connecting to server
    ConnectSocket=socket(ptr->ai_family,ptr->ai_socktype,ptr->ai_protocol);
    if (ConnectSocket==INVALID_SOCKET) {
      printf("socket failed with error: %d\n", WSAGetLastError());
      WSACleanup();
      return 1;
    }

    // Connect to server.
    r=connect(ConnectSocket,ptr->ai_addr,(int)ptr->ai_addrlen);
    if (r==SOCKET_ERROR) {
      closesocket(ConnectSocket);
      ConnectSocket=INVALID_SOCKET;
      continue;
    }
    break;
  }
  fprintf(stderr,"\n");
  
  freeaddrinfo(result);
  
  if (ConnectSocket==INVALID_SOCKET) {
    printf("Unable to connect to server!\n");
    WSACleanup();
    return 1;
  }

  Sleep(2000);

  reset_data_packet(&data);
  for (count=0,done=0;!done;count++) {
    data.ctrl = 0;
    switch(CtrlIF) {
    case 0: // Keyboard
      printf("enter a control character: ");
      do {
	input = getchar();
      } while (input<=' ');
      switch (input) {
      case 'q': case 'Q': case 'x': case 'X':
	done=1;
	break;
      case 's':
	data.servoA += 10000;
	break;
      case 'S':
	data.servoA -= 10000;
	break;
      case 't':
	data.servoB += 10000;
	break;
      case 'T':
	data.servoB -= 10000;
	break;
      case 'b': case 'B':
	data.ctrl = MOTOR_BCKL|MOTOR_BCKR;
	data.pwmA=MOTOR_LOW;
	data.pwmB=MOTOR_LOW;
	sleeptime=500;
	break;
      case 'f': case 'F':
	data.ctrl = MOTOR_FWDL|MOTOR_FWDR;
	data.pwmA=MOTOR_LOW;
	data.pwmB=MOTOR_LOW;
	sleeptime=500;
	break;
      case 'l': case 'L':
	data.ctrl = MOTOR_BCKL|MOTOR_FWDR;
	data.pwmA=MOTOR_LOW;
	data.pwmB=MOTOR_LOW;
	sleeptime=500;
	break;
      case 'r': case 'R':
	data.ctrl = MOTOR_BCKR|MOTOR_FWDL;
	data.pwmA=MOTOR_LOW;
	data.pwmB=MOTOR_LOW;
	sleeptime=500;
	break;
      case 'j': case 'J':
	if(JOYERR_NOERROR==joyGetPosEx(0,&joypad)) {
	  print_help();
	  printf("Controlled by Joypad\n");
	  CtrlIF=1;
	}
	break;
      default:
	break;
      }

      print_data_packet(&data);
      r=send(ConnectSocket,(char *)(&data),sizeof(data),0);
      if (r==SOCKET_ERROR) {
	printf("send failed with error: %d\n", WSAGetLastError());
	closesocket(ConnectSocket);
	WSACleanup();
	return 1;
      }

      Sleep(sleeptime);
      data.ctrl=0;// sent later

      break; // case Keyboard
    case 1: // JoyPad
      if(JOYERR_NOERROR==joyGetPosEx(0,&joypad)){
	if (joypad.dwButtons&0x004) data.servoA += 2000;
	if (joypad.dwButtons&0x002) data.servoA -= 2000;
	if (joypad.dwButtons&0x001) data.servoB += 2000;
	if (joypad.dwButtons&0x008) data.servoB -= 2000;
	// left wheel
	joypad_R=joypad.dwRpos-joypad_R0;
	if (joypad_R>JoypadMargin) {
	  ratio=(joypad_R-JoypadMargin)/(MOTOR_MAX/2.0-JoypadMargin);
	  if (ratio>=1.0) ratio=1.0;
	  data.pwmA=(MOTOR_MAX-MOTOR_LOW)*ratio+MOTOR_LOW;
	  data.ctrl |= MOTOR_BCKL;
	} else if (joypad_R<-JoypadMargin) {
	  ratio=(-joypad_R-JoypadMargin)/(MOTOR_MAX/2.0-JoypadMargin);
	  if (ratio>=1.0) ratio=1.0;
	  data.pwmA=(MOTOR_MAX-MOTOR_LOW)*ratio+MOTOR_LOW;
	  data.ctrl |= MOTOR_FWDL;
	} else {
	  data.pwmA=0;
	}
	// right wheel
	joypad_Y=joypad.dwYpos-joypad_Y0;
	if (joypad_Y>JoypadMargin) {
	  ratio=(joypad_Y-JoypadMargin)/(MOTOR_MAX/2.0-JoypadMargin);
	  if (ratio>=1.0) ratio=1.0;
	  data.pwmB=(MOTOR_MAX-MOTOR_LOW)*ratio+MOTOR_LOW;
	  data.ctrl |= MOTOR_BCKR;
	} else if (joypad_Y<-JoypadMargin) {
	  ratio=(-joypad_Y-JoypadMargin)/(MOTOR_MAX/2.0-JoypadMargin);
	  if (ratio>=1.0) ratio=1.0;
	  data.pwmB=(MOTOR_MAX-MOTOR_LOW)*ratio+MOTOR_LOW;
	  data.ctrl |= MOTOR_FWDR;
	} else {
	  data.pwmB=0;
	}
	if (joypad.dwButtons&0x010) {
	  print_help();
	  printf("Controlled by Keyboard\n");
	  CtrlIF=0;
	}
	if (joypad.dwButtons&0x020) done=1;
	if (joypad.dwButtonNumber>3) done=1;
      } else { // Joypad trouble;
	done = 1;
      }
      print_data_packet(&data);
      break;
    default:
      done = 1;
    }

    r=send(ConnectSocket,(char *)(&data),sizeof(data),0);
    if (r==SOCKET_ERROR) {
      printf("send failed with error: %d\n",WSAGetLastError());
      closesocket(ConnectSocket);
      WSACleanup();
      return 1;
    }
    
    Sleep(100);
  }

  // cleanup
  closesocket(ConnectSocket);
  WSACleanup();
  
  return 0;
}
