/*
  access to User I/O device
  by izumi@ieee.org ver.2016/10/07a
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <fcntl.h>

int main(int argc,char *argv[]){
  int fd;
  int dirout=0;
  unsigned int value=0;
  char *uioname;
  unsigned int offset=0;
  unsigned page_size=sysconf(_SC_PAGESIZE);
  void *ptr;

  if (argc==4) {
    /* output */
    dirout=1;
    uioname=argv[1];
    offset=strtoul(argv[2],NULL,0);
    value=strtoul(argv[3],NULL,0);
  } else if (argc==3) {
    /* input */
    dirout=0;
    uioname=argv[1];
    offset=strtoul(argv[2],NULL,0);
  } else if (argc==2) {
    /* input */
    dirout=0;
    uioname=argv[1];
    offset=0;
  } else {
    printf("IO access through /dev/uio*.\n");
    printf("%s <DEVFILE> <OFFSET> [<VALUE>]\n",argv[0]);
    printf("	<DEVFILE>  UIO device file. e.g. /dev/uio0\n");
    printf("	<OFFSET>   offset of address\n");
    printf("	<VALUE>    output value to IO\n");
    return -1;
  }

  if (offset<0||offset>page_size) {
    fprintf(stderr,"offset is out of range.\n");
    return -1;
  }

  /* Open the UIO device file */
  fd=open(uioname,O_RDWR);
  if (fd<=0) {
    fprintf(stderr,"can't open UIO device file:%s.\n",uioname);
    return -1;
  }

  /* mmap the UIO device */
  ptr=mmap(NULL,page_size,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
  if (ptr==0) {
    fprintf(stderr,"failed to mmap the UIO device.\n");
    return -1;
  }

  if (dirout) {
    /* write to IO */
    *((unsigned int *)(ptr+offset))=value;
    printf("%s[%d] <= %08x\n",uioname,offset,value);
  } else {
    /* read from IO */
    value= *((unsigned int *)(ptr+offset));
    printf("%s[%d] => %08x\n",uioname,offset,value);
  }

  munmap(ptr,page_size);
  close(fd);	

  return 0;
}
